/** \file     pi1.c
 *  \brief   Prime Implicant Computation Based on MetaProduct Encoding
 *  \author  Takahisa Toda
 *  \note    see O.Coudert and J.C. Madre, Implicit and Incremental Computation of Primes and Essential Primes of Boolean functions, 29th ACM/IEEE Design Automation Conference, 1992.
 *  \note    see I.Wegener. Branching Programs and Binary Decision Diagrams: Theory and Applications. SIAM Monographs on Discrete Mathematics and Applications, 2000. 
 */


#include "pi1.h"
#include "my_hash.h"

static zddp pi1_rec(bddp f, my_hash *h);

static uintmax_t recdepth = 0; //!< recursion depth


/** \brief    Construct the ZDD that represents all prime implicants for the Boolean function represented by a BDD.
 *  \param    f   ZDD
 *  \return   The ZDD contructed.
 *  \note     Variables in the resulted ZDD consists of occurence variables and sign variables according to MetaProduct encoding.
 */
zddp pi1(bddp f)
{
  my_hash *h = ht_create(0);
  ENSURE_TRUE_MSG(h != NULL, "hash table creation failed");

  assert(recdepth == 0);
  zddp r = pi1_rec(f, h);
  ENSURE_TRUE(r != BDD_NULL);
  assert(recdepth == 0);

  ht_destroy(h);
  return r;
}

static zddp pi1_rec(bddp f, my_hash *h)
{
  if(f == bdd_top()) return zdd_top();
  if(f == bdd_bot()) return zdd_bot();

  zddp r;
  if(ht_search((uintptr_t)f, (uintptr_t*)&r, h)) return r;

  INC_RECDEPTH(recdepth);
  const int v = zdd_itemval(f);
  zddp r0 = pi1_rec(bdd_and(bdd_lo(f),bdd_hi(f)), h);  
  zddp r1 = zdd_diff(pi1_rec(bdd_lo(f), h), r0);
  ENSURE_TRUE_MSG(r1 != BDD_NULL, "BDD operation failed");
  zddp r2 = zdd_diff(pi1_rec(bdd_hi(f), h), r0);
  ENSURE_TRUE_MSG(r2 != BDD_NULL, "BDD operation failed");
  DEC_RECDEPTH(recdepth);

  zddp t = zdd_node(2*v, r1, r2);
  r = zdd_node(2*v-1, r0, t);
  ENSURE_TRUE_MSG(r != BDD_NULL, "BDD operation failed");

  ht_insert((uintptr_t)f, (uintptr_t)r, h);
  
  return r;
}
