/** \file    pi2.c
 *  \brief   Prime Implicant Computation Based on Polarity Encoding
 *  \author  Takahisa Toda
 *  \note    see S. Minato. Zero-Suppressed BDDs and Their Applications. International Journal on Software Tools for Technology Transfer, 3(2): 156-170,2001.
 *  \note    see A. Mischenko. An Introduction to Zero-Suppressed Binary Decision Diagrams. Technical report, Portland State University, June 2001.
 */


#include "pi2.h"
#include "my_hash.h"

static zddp pi2_rec(bddp f, my_hash *h);

static uintmax_t recdepth = 0; //!< recursion depth


/** \brief    Construct the ZDD that represents all prime implicants for the Boolean function represented by a BDD.
 *  \param    f   ZDD
 *  \return   The ZDD contructed.
 *  \note     Variables in the resulted ZDD consists of positive variables and negative variables according to positive and negative bits encoding.
 */
zddp pi2(bddp f)
{
  my_hash *h = ht_create(0);
  ENSURE_TRUE_MSG(h != NULL, "hash table creation failed");

  assert(recdepth == 0);
  zddp r = pi2_rec(f, h);
  ENSURE_TRUE(r != BDD_NULL);
  assert(recdepth == 0);

  ht_destroy(h);
  return r;
}

static zddp pi2_rec(bddp f, my_hash *h)
{
  if(f == bdd_top()) return zdd_top();
  if(f == bdd_bot()) return zdd_bot();

  zddp r;
  if(ht_search((uintptr_t)f, (uintptr_t*)&r, h)) return r;

  INC_RECDEPTH(recdepth);
  const itemval v = zdd_itemval(f);
  zddp r0 = pi2_rec(bdd_and(bdd_lo(f),bdd_hi(f)), h);  
  zddp r1 = zdd_diff(pi2_rec(bdd_lo(f), h), r0);
  ENSURE_TRUE_MSG(r1 != BDD_NULL, "BDD operation failed");
  zddp r2 = zdd_diff(pi2_rec(bdd_hi(f), h), r0);
  ENSURE_TRUE_MSG(r2 != BDD_NULL, "BDD operation failed");
  DEC_RECDEPTH(recdepth);

  zddp t1 = zdd_node(2*v, r0, r1);
  zddp t2 = zdd_node(2*v, r2, zdd_bot());
  r = zdd_node(2*v-1, t1, t2);
  ENSURE_TRUE_MSG(r != BDD_NULL, "BDD operation failed");

  ht_insert((uintptr_t)f, (uintptr_t)r, h);
  
  return r;
}

